import symengine

class GroupHandler(object):
	"""
	Class to handle groups of synchronised variables for transversal Lyapunov exponents. Main indices are those that represent the normal dynamics, tangent indices are those that belong to tangent vectors. See the accompanying paper for the mathematical background.
	"""
	
	def __init__(self,groups):
		flattened_groups = [i for group in groups for i in group]
		self.n = max(flattened_groups)+1
		assert all(0<=i<self.n for i in flattened_groups), "Group elements out of range."
		assert set(flattened_groups)==set(range(self.n)), "Groups do not cover all indices. If you do not want a dynamical variable to be part of a synchronisation group, add it to a group of its own."
		assert len(flattened_groups)==self.n, "Groups overlap."
		
		self.groups = [sorted(group) for group in groups]
	
	@property
	def main_indices(self):
		if not hasattr(self,"_main_indices"):
			self._main_indices = [group[0] for group in self.groups]
		return self._main_indices
	
	@property
	def tangent_indices(self):
		if not hasattr(self,"_tangent_indices"):
			self._tangent_indices = sorted([
					i
					for group in self.groups
					for i in group[1:]
				])
		return self._tangent_indices
	
	def group_from_index(self,index):
		"""returns the number of the group which contains `i`."""
		
		if not hasattr(self,"_group_finder"):
			self._group_finder = {}
			for k,group in enumerate(self.groups):
				for entry in group:
					self._group_finder[entry] = k
		
		return self._group_finder[index]
	
	def map_to_main(self,index):
		return self.main_indices[self.group_from_index(index)]
	
	def iterate(self,iterable):
		"""
		For a main index, return the number of the respective group.
		For a tangent index, returns the respective element and the preceding one in the same group.
		"""
		
		cache = {}
		for k,entry in enumerate(iterable):
			i = self.group_from_index(k)
			try:
				previous_entry = cache[i]
			except KeyError:
				yield i
			else:
				yield previous_entry,entry
			finally:
				cache[i] = entry
	
	def extract_main(self,generator_function):
		"""
		Extracts the elements corresponding to the main indices from `generator_function`.

		Returns
		-------
		extractor: generator function
			returns an iterator over iterable
		
		extracted_entries : dictionary
			maps an index of a main component to the corresponding element of `iterable`. Note that this only works once `extractor` has been iterated sufficienly far.
		"""
		
		extracted_entries = {}
		def extractor():
			for i,entry in enumerate(generator_function()):
				if i in self.main_indices:
					extracted_entries[i] = entry
				yield entry
		return extractor,extracted_entries
	
	def back_transform(self,vector):
		result = [0]*self.n
		for group in self.groups:
			# Uppercase numbers are indices of the respective submatrix, lowercase numbers are indices of the full matrix
			
			N = symengine.Integer(len(group))
			for I,i in enumerate(group):
				for J in range(1,I+1):
					j = group[J]
					result[i] +=   -J/N *vector[j]
				for J in range(I+1,len(group)):
					j = group[J]
					result[i] += (1-J/N)*vector[j]
		
		return result
		



#!/usr/bin/python3
# -*- coding: utf-8 -*-

from jitcxde_common.helpers import sort_helpers, filter_helpers, copy_helpers, find_dependent_helpers
import unittest
from symengine import symbols, sin, cos, Integer
from itertools import permutations

p,q,r,s,u,v = symbols("p q r s u v")

cycle = [ [p,q], [q,r], [r,p] ]
chain = [ [r,s], [q,r], [p,q] ]

class SortingTest(unittest.TestCase):
	def test_cyclic(self):
		for permutation in permutations(cycle):
			with self.assertRaises(ValueError):
				sort_helpers(permutation)
	
	def test_sorting(self):
		for permutation in permutations(chain):
			result = sort_helpers(list(permutation))
			assert result==chain

class FilterTest(unittest.TestCase):
	def test_all_needed(self):
		assert chain==filter_helpers(chain,{p})
	
	def test_spurious(self):
		chain_with_spurious = chain+[[u,v]]
		assert chain==filter_helpers(chain_with_spurious,{p})

class CopyTest(unittest.TestCase):
	def test_copy(self):
		assert chain==copy_helpers(chain)
	
	def test_independence(self):
		copy = copy_helpers(chain)
		copy.append([u,v])
		assert copy!=chain

class FindDependentTest(unittest.TestCase):
	def test_find_dependent_helpers(self):
		helpers = [
				( q, p           ),
				( r, sin(q)      ),
				( s, 3*p+r       ),
				( u, Integer(42) ),
			]
		control = [
				( q, 1        ),
				( r, cos(q)   ),
				( s, 3+cos(q) ),
			]
		dependent_helpers = find_dependent_helpers(helpers,p)
		# This check is overly restrictive due to depending on the order and exact form of the result:
		self.assertListEqual(dependent_helpers,control)

if __name__ == "__main__":
	unittest.main(buffer=True)





# This creates an infrastructure for checks for JiTC*DE objects that works as follows:
# • Each check is registered as such by being decorated with `@checker`.
# • Each class that needs checks inherits from CheckEnvironment.
# • All checks are run with the method check.

from functools import update_wrapper

# This class exist just to mark functions
class checker(object):
	def __init__(self,function):
		self.function = function
		update_wrapper(self,function)
	
	def __call__(self,*args):
		self.function(*args)

class CheckEnvironment(object):
	def _check_assert(self,condition,message):
		if not condition:
			self.failed_check = True
			if self.fail_checks_fast:
				raise ValueError(message)
			else:
				print(message)
	
	def check(self, fail_fast=True):
		"""
		Performs a series of checks that may not be feasible at runtime (usually due to their length). Whenever you run into an error that you cannot make sense of, try running this. It checks for the following mistakes:
		
		* negative arguments of `y`
		* arguments of `y` that are higher than the system’s dimension `n`
		* unused variables
		
		Parameters
		----------
		fail_fast : boolean
			whether to abort on the first failure. If false, an error is raised only after all problems are printed.
		"""
		
		self.failed_check = False
		self.fail_checks_fast = fail_fast
		
		# execute all methods decorated with checker:
		visited = set()
		for cls in [self.__class__] + self.__class__.mro():
			for name,member in cls.__dict__.items():
				if name not in visited and isinstance(member,checker):
					member(self)
				visited.add(name)
		
		if self.failed_check:
			raise ValueError("Check failed.")




#!/usr/bin/python3
# -*- coding: utf-8 -*-

import numpy as np
from numpy.testing import assert_allclose
import unittest
from jitcxde_common.numerical import random_direction, orthonormalise, rel_dist

class RandomDirectionTest(unittest.TestCase):
	def test_random_direction(self):
		d = 10
		n = 100000
		n_vectors = np.vstack([random_direction(d) for i in range(n)])
		average = np.average(n_vectors, axis=0)
		assert_allclose( average, np.zeros(d), rtol=0, atol=0.01 )

class OrthonormaliseTest(unittest.TestCase):
	def test_orthonormalise_1(self):
		vectors = [ np.array([3.0,4.0]) ]
		norms = orthonormalise(vectors)
		assert_allclose( vectors[0], np.array([0.6,0.8]) )
		assert_allclose( norms, np.array([5]) )
	
	def test_orthonormalise_2(self):
		vectors = [ np.array([1.0,0.0]), np.array([0.0,1.0]) ]
		norms = orthonormalise(vectors)
		assert_allclose( vectors[0], np.array([1.0,0.0]) )
		assert_allclose( vectors[1], np.array([0.0,1.0]) )
		assert_allclose( norms, np.array([1.0,1.0]) )
	
	def test_orthonormalise_3(self):
		vectors = [ np.array([1.0,0.0]), np.array([1.0,1.0]) ]
		norms = orthonormalise(vectors)
		assert_allclose( vectors[0], np.array([1.0,0.0]) )
		assert_allclose( vectors[1], np.array([0.0,1.0]) )
		assert_allclose( norms, np.array([1.0,1.0]) )
	
	def test_orthonormalise_4(self):
		vectors = [ np.array([1.0,1.0]), np.array([1.0,1.0]) ]
		norms = orthonormalise(vectors)
		assert_allclose( vectors[0], np.array([np.sqrt(0.5),np.sqrt(0.5)]) )
		assert_allclose( norms, np.array([np.sqrt(2),0.0]) , atol = 1e-10 )
	
	def test_orthonormalise_5(self):
		vectors = [ np.array([1.0,1.0]), np.array([1.0,-1.0]) ]
		norms = orthonormalise(vectors)
		assert_allclose( vectors[0], np.array([np.sqrt(0.5),np.sqrt(0.5)]) )
		assert_allclose( vectors[1], np.array([np.sqrt(0.5),-np.sqrt(0.5)]) )
		assert_allclose( norms, np.array([np.sqrt(2),np.sqrt(2)]) )
	
	def test_orthonormalise_6(self):
		vectors = [ np.array([1.0,1.0]), np.array([1.0,0.0]) ]
		norms = orthonormalise(vectors)
		assert_allclose( vectors[0], np.array([np.sqrt(0.5),np.sqrt(0.5)]) )
		assert_allclose( vectors[1], np.array([np.sqrt(0.5),-np.sqrt(0.5)]) )
		assert_allclose( norms, np.array([np.sqrt(2),np.sqrt(0.5)]) )

class RelDistTest(unittest.TestCase):
	def test_reldist_1(self):
		assert_allclose( rel_dist(1,3), 1.0 )
	
	def test_reldist_2(self):
		assert_allclose( rel_dist(1,1+1e-10), 1e-10 )
	
	def test_reldist_3(self):
		assert_allclose( rel_dist([1,1],[1+1e-10,1]), 1e-10 )
	
if __name__ == "__main__":
	unittest.main(buffer=True)



def rsplit_int(s):
	"""
	splits off the largest substring of digits from the right a string
	"""
	if s and s[-1].isdigit():
		x,y = rsplit_int(s[:-1])
		return x, y+s[-1]
	else:
		return s, ""

def count_up(name):
	"""
	If `name` ends on a number, increase that number by one. Otherwise append “_1”.
	"""
	s, i = rsplit_int(name)
	return s + ( "%%.%ii" % len(i) % (int(i)+1)  if i else "_1" )




#!/usr/bin/python3
# -*- coding: utf-8 -*-

import os
import platform
import shutil
from tempfile import mkdtemp
import unittest
import pickle

import symengine
import numpy

from jitcxde_common import jitcxde

y = symengine.Function("y")
f = [
		y(0),
		2*y(1)**3,
		y(0)+y(1)+y(2),
		symengine.exp(y(3)),
		5,
	]

def f_control(y):
	return [
		y[0],
		2*y[1]**3,
		y[0]+y[1]+y[2],
		numpy.exp(y[3]),
		5,
	]

def f_generator():
	for entry in f:
		yield entry

f_dictionary = { y(i):entry for i,entry in enumerate(f) }

class jitcxde_tester(jitcxde):
	dynvar = y
	
	def __init__(self,f_sym=(),n=None,module_location=None,chunk_size=100,omp=False):
		jitcxde.__init__(self,n,False,module_location)
		self.f_sym = self._handle_input(f_sym)
		set_dy = symengine.Function("set_dy")
		self.omp = omp
		
		self.render_and_write_code(
			(set_dy(i,entry) for i,entry in enumerate(self.f_sym())),
			name = "f",
			chunk_size = chunk_size,
			arguments = [
					("Y" , "PyArrayObject *__restrict const"),
					("dY", "PyArrayObject *__restrict const")
				]
			)
	
	def _compile_C(self):
		if self.jitced is None:
			self.compile_C()
	
	def compile_C(self,modulename=None):
		self._process_modulename(modulename)
		self._render_template(n=self.n)
		self._compile_and_load(False,None,None,self.omp)


name = ""
def get_unique_name():
	global name
	name += "x"
	return name

class basic_test(unittest.TestCase):
	def tmpfile(self, filename):
		return os.path.join(self.directory, filename)
	
	@classmethod
	def setUpClass(self):
		self.argdict = {"f_sym": f}
	
	def setUp(self):
		self.directory = mkdtemp()
		self.tester = jitcxde_tester(**self.argdict)
	
	def test_default(self):
		self.tester.compile_C()
	
	def test_save_and_load(self):
		destination = self.tester.save_compiled(overwrite=True)
		folder, filename = os.path.split(destination)
		shutil.move(filename,self.tmpfile(filename))
		self.tester = jitcxde_tester(module_location=self.tmpfile(filename))
	
	def test_compile_save_and_load(self,default=False):
		modulename = None if default else get_unique_name()
		self.tester.compile_C(modulename=modulename)
		filename = self.tester.save_compiled(overwrite=True)
		shutil.move(filename, self.tmpfile(filename))
		self.tester = jitcxde_tester(module_location=self.tmpfile(filename))
	
	def test_save_with_default_name_and_load(self):
		self.test_compile_save_and_load(True)
	
	def test_save_to_directory_and_load(self):
		modulename = get_unique_name()
		self.tester.compile_C(modulename=modulename)
		destination = self.tester.save_compiled(self.tmpfile(""), overwrite=True)
		folder, filename = os.path.split(destination)
		print(folder,self.tmpfile(""))
		assert folder==os.path.dirname(self.tmpfile(""))
		self.tester = jitcxde_tester(module_location=self.tmpfile(filename))
	
	def tearDown(self):
		arg = numpy.random.uniform(-2,2,5)
		numpy.testing.assert_allclose(
				self.tester.jitced.f(arg),
				f_control(arg)
			)
		
		if platform.system() != "Windows":
			# Windows blocks loaded module files from removal.
			shutil.rmtree(self.directory)

class basic_test_with_chunking(basic_test):
	@classmethod
	def setUpClass(self):
		self.argdict = {"f_sym":f, "chunk_size":1}

class basic_test_with_omp(basic_test):
	@classmethod
	def setUpClass(self):
		self.argdict = {"f_sym":f, "chunk_size":1, "omp":True}

class basic_test_with_generator_function(basic_test):
	@classmethod
	def setUpClass(self):
		self.argdict = {"f_sym":f_generator, "n":len(f)}

class basic_test_with_dictionary(basic_test):
	@classmethod
	def setUpClass(self):
		self.argdict = {"f_sym":f_dictionary}

class test_errors(unittest.TestCase):
	def test_not_supported_error(self):
		x = symengine.Symbol("x")
		faulty_f = [y(x).diff(x)]
		with self.assertRaises(NotImplementedError):
			jitcxde_tester(faulty_f)
	
	def test_pickle_error(self):
		tester = jitcxde_tester(f)
		with self.assertRaises(pickle.PickleError):
			pickle.dumps(tester)
	
	def test_dict_missing_equation(self):
		faulty_f = { y(0):1, y(2):1 }
		with self.assertRaises(ValueError):
			jitcxde_tester(faulty_f)
	
	def test_dict_spurious_equation(self):
		x = symengine.Symbol("x")
		faulty_f = { y(0):1, y(1):1, x:1 }
		with self.assertRaises(ValueError):
			jitcxde_tester(faulty_f)
	
	def test_dict_tester_missing_equation(self):
		tester = jitcxde_tester(f)
		faulty_dict = { y(0):1, y(2):1 }
		for i in range(10):
			with self.assertRaises(ValueError):
				tester._check_dynvar_dict(faulty_dict,"",i)
	
	def test_dict_spurious_equation(self):
		tester = jitcxde_tester(f)
		x = symengine.Symbol("x")
		faulty_dict = { y(0):1, y(1):1, x:1 }
		for i in range(10):
			with self.assertRaises(ValueError):
				tester._check_dynvar_dict(faulty_dict,"",i)
	
	def test_set(self):
		with self.assertRaises(ValueError):
			jitcxde_tester(set(f))
	
	def test_too_generator_length_mismatch(self):
		for mismatch in [-1,1]:
			X = jitcxde_tester(f_generator,n=len(f)+mismatch)
			with self.assertRaises(ValueError):
				X.check()

if __name__ == "__main__":
	unittest.main(buffer=True)




from os import path
from importlib.util import module_from_spec
from importlib.machinery import ExtensionFileLoader, EXTENSION_SUFFIXES, FileFinder
from importlib.util import spec_from_file_location

loader_details = (ExtensionFileLoader, EXTENSION_SUFFIXES)
suffices = sorted(EXTENSION_SUFFIXES, key=len)

def remove_suffix(path):
	for suffix in reversed(suffices):
		if path.endswith(suffix):
			return path.rpartition(suffix)[0]
	else:
		return path

def add_suffix(path):
	for suffix in suffices:
		if path.endswith(suffix):
			return path
	else:
		return path + suffices[0]

def modulename_from_path(full_path):
	return remove_suffix(path.basename(full_path))

def get_module_path(modulename, folder=""):
	finder = FileFinder(folder, loader_details)
	return finder.find_spec(modulename).origin

def find_and_load_module(modulename, folder=""):
	finder = FileFinder(folder, loader_details)
	spec = finder.find_spec(modulename)
	module = module_from_spec(spec)
	spec.loader.exec_module(module)
	return module

def module_from_path(full_path):
	modulename = modulename_from_path(full_path)
	spec = spec_from_file_location(modulename, full_path)
	module = module_from_spec(spec)
	spec.loader.exec_module(module)
	return module




#!/usr/bin/python3
# -*- coding: utf-8 -*-

import os
import platform
import shutil
from tempfile import mkdtemp
import unittest
import pickle

import symengine
import numpy

from jitcxde_common import jitcxde

y = symengine.Function("y")
f = [
		y(0),
		2*y(1)**3,
		y(0)+y(1)+y(2),
		symengine.exp(y(3)),
		5,
	]

def f_control(y):
	return [
		y[0],
		2*y[1]**3,
		y[0]+y[1]+y[2],
		numpy.exp(y[3]),
		5,
	]

class jitcxde_tester(jitcxde):
	dynvar = y
	
	def __init__(self,f_sym=(),omp=False):
		jitcxde.__init__(self,len(f),False)
		f_sym_wc = self._handle_input(f_sym)
		set_dy = symengine.Function("set_dy")
		self.omp = omp
		
		self.render_and_write_code(
			(set_dy(i,entry) for i,entry in enumerate(f_sym_wc())),
			name = "f",
			chunk_size = 1 if omp else 100,
			arguments = [
					("Y" , "PyArrayObject *__restrict const"),
					("dY", "PyArrayObject *__restrict const")
				]
			)
	
	def _compile_C(self):
		if self.jitced is None:
			self.compile_C()
	
	def compile_C(self,modulename=None):
		self._process_modulename(modulename)
		self._render_template(n=self.n)
		self._compile_and_load(False,None,None,self.omp)


name = ""
def get_unique_name():
	global name
	name += "x"
	return name

for omp in [True, False]:
	tester = jitcxde_tester(f,omp)
	tester.compile_C()
	arg = numpy.random.uniform(-2,2,5)
	numpy.testing.assert_allclose(
			tester.jitced.f(arg),
			f_control(arg)
		)






import unittest
import symengine
from jitcxde_common.transversal import GroupHandler

class TestOrdered(unittest.TestCase):
	@classmethod
	def setUpClass(self):
		self.n = 6
		self.groups = [ [0,1,2], [3,4,5] ]
		self.main_indices = [0,3]
		self.tangent_indices = [1,2,4,5]
		self.iteration = [ 0 , (0,1), (1,2), 1 , (3,4), (4,5) ]
	
	def setUp(self):
		self.G = GroupHandler(self.groups)
	
	def test_group_from_index(self):
		for i in range(self.n):
			self.assertIn(i,self.groups[self.G.group_from_index(i)])
	
	def test_iterate(self):
		self.assertSequenceEqual(
				list(self.G.iterate(range(self.n))),
				self.iteration
			)
		
	def test_main_indices(self):
		self.assertSequenceEqual(
				list(self.G.main_indices),
				self.main_indices
			)
	
	def test_tangent_indices(self):
		self.assertCountEqual(
				list(self.G.tangent_indices),
				self.tangent_indices
			)
	
	def test_extractor(self):
		original = range(100,100+self.n)
		extractor,extracted = self.G.extract_main(lambda:original)
		sequence = list(extractor())
		self.assertSequenceEqual(sequence,original)
		for i in self.G.main_indices:
			self.assertEqual(extracted[i],original[i])
	
	def test_back_transform(self):
		z = symengine.Function("z")
		
		z_v = [z(i) for i in range(self.n)]
		y_v = self.G.back_transform(z_v)
		
		transformed = []
		for entry in self.G.iterate(range(self.n)):
			if type(entry)==int:
				transformed.append( sum(y_v[i] for i in self.groups[entry]) )
			else:
				transformed.append( y_v[entry[0]] - y_v[entry[1]] )
		
		for i in self.G.main_indices:
			z_v[i] = 0
		self.assertSequenceEqual(
				z_v,
				[entry.simplify() for entry in transformed]
			)

class TestAlternating(TestOrdered):
	@classmethod
	def setUpClass(self):
		self.n = 6
		self.groups = [ [0,2,4], [1,3,5] ]
		self.main_indices = [0,1]
		self.tangent_indices = [2,3,4,5]
		self.iteration = [ 0 , 1 , (0,2), (1,3), (2,4), (3,5) ]

class TestErrors(unittest.TestCase):
	def test_missing_indices(self):
		groups = [(0,1),(3,4)]
		with self.assertRaises(AssertionError):
			GroupHandler(groups)
	
	def test_duplicte_indices(self):
		groups = [(0,1),(1,2,3,4)]
		with self.assertRaises(AssertionError):
			GroupHandler(groups)

if __name__ == "__main__":
	unittest.main(buffer=True)

